
/* (c) 2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include "defines.h"
#include "mesh.h"
#include "parse.h"



struct object_mesh *mesh_create(void) {

  struct object_mesh *m;


  m = malloc(sizeof(struct object_mesh));
  if (m == NULL) {
    fprintf(stderr, "MESH_CREATE: Out of memory error.\n");
    return NULL;
  }

  m->name[0] = 0;
  m->vertices = 0;
  m->triangles = 0;
  m->collision_status = 0;
  m->triangle_list = NULL;
  m->textures = NULL;
  m->triangle_cube = NULL;
  m->triangle_center_xyz = NULL;
  m->vertice_list = NULL;
  m->vertex_normal_list = NULL;
  m->face_normal_list = NULL;
  m->uv_list = NULL;
  m->triangle_radius_list = NULL;
  m->next = NULL;

  return m;
}


int mesh_free(struct object_mesh *mesh) {

  if (mesh == NULL)
    return FAILED;

  /* triangle_list & textures */
  if (mesh->textures != NULL)
    free(mesh->textures);

  /* vertex_normal_list & face_normal_list */
  if (mesh->vertex_normal_list != NULL)
    free(mesh->vertex_normal_list);

  if (mesh->vertice_list != NULL)
    free(mesh->vertice_list);
  if (mesh->uv_list != NULL)
    free(mesh->uv_list);
  if (mesh->triangle_radius_list != NULL)
    free(mesh->triangle_radius_list);

  if (mesh->triangle_cube != NULL)
    free(mesh->triangle_cube);
  if (mesh->triangle_center_xyz != NULL)
    free(mesh->triangle_center_xyz);

  free(mesh);

  return SUCCEEDED;
}


int mesh_write(struct object_mesh *m, FILE *f) {

  double *d;
  float *l;
  int no, x, *z;


  if (m == NULL)
    return FAILED;

  if (m->vertex_normal_list != NULL && m->face_normal_list != NULL)
    no = YES;
  else
    no = NO;

  /* write mesh header */
  fprintf(f, "// Mesh data\n");
  fprintf(f, "\"%s\"\n", m->name);
  fprintf(f, "Vertices: %d\n", m->vertices);
  fprintf(f, "Normals: ");
  if (no == YES)
    fprintf(f, "YES\n");
  else
    fprintf(f, "NO\n");
  fprintf(f, "Triangles: %d\n", m->triangles);
  fprintf(f, "// Vertice data\n");

  /* vertices */
  d = m->vertice_list;
  for (x = 0; x < m->vertices; x++) {
    write_out_double(f, *(d+0), ' ');
    write_out_double(f, *(d+1), ' ');
    write_out_double(f, *(d+2), 0xA);
    d += 3;
  }

  /* vertex normals */
  if (no == YES) {
    fprintf(f, "// Vertex normals\n");
    l = m->vertex_normal_list;
    for (x = 0; x < m->vertices; x++) {
      write_out_double(f, *(l+0), ' ');
      write_out_double(f, *(l+1), ' ');
      write_out_double(f, *(l+2), 0xA);
      l += 3;
    }
  }

  /* face normals */
  if (no == YES) {
    fprintf(f, "// Face normals\n");
    l = m->face_normal_list;
    for (x = 0; x < m->triangles; x++) {
      write_out_double(f, *(l+0), ' ');
      write_out_double(f, *(l+1), ' ');
      write_out_double(f, *(l+2), 0xA);
      l += 3;
    }
  }

  /* triangles */
  fprintf(f, "// Triangle data\n");
  z = m->triangle_list;
  l = m->uv_list;
  for (x = 0; x < m->triangles; x++) {
    fprintf(f, "%d %d %d %d ", m->textures[x], z[0], z[1], z[2]);
    write_out_double(f, *(l+0), ' ');
    write_out_double(f, *(l+1), ' ');
    write_out_double(f, *(l+2), ' ');
    write_out_double(f, *(l+3), ' ');
    write_out_double(f, *(l+4), ' ');
    write_out_double(f, *(l+5), 0xA);
    z += 3;
    l += 6;
  }

  return SUCCEEDED;
}


int mesh_compute_triangle_bounding_spheres(struct object_mesh *mesh, int root) {

  double dx1, dx2, dy1, dy2, dz1, dz2, x1, x2, x3, y1, y2, y3, z1, z2, z3, *t;
  float *f;
  int i, *l, n;


  if (mesh == NULL)
    return FAILED;

  if (mesh->triangle_radius_list == NULL) {
    mesh->triangle_radius_list = malloc(sizeof(float) * mesh->triangles);
    if (mesh->triangle_radius_list == NULL) {
      fprintf(stderr, "MESH_COMPUTE_TRIANGLE_BOUNDING_SPHERES: Out of memory error.\n");
      return FAILED;
    }
  }

  /* estimate triangles with spheres, compute radius for such a sphere that eats the whole triangle */
  l = mesh->triangle_list;
  f = mesh->triangle_radius_list;
  t = mesh->vertice_list;
  for (i = 0; i < mesh->triangles; i++) {
    n = *(l++) * 3;
    x1 = t[n + 0];
    y1 = t[n + 1];
    z1 = t[n + 2];
    n = *(l++) * 3;
    x2 = t[n + 0];
    y2 = t[n + 1];
    z2 = t[n + 2];
    n = *(l++) * 3;
    x3 = t[n + 0];
    y3 = t[n + 1];
    z3 = t[n + 2];

    dx1 = x1 - x2;
    dy1 = y1 - y2;
    dz1 = z1 - z2;
    dx2 = x1 - x3;
    dy2 = y1 - y3;
    dz2 = z1 - z3;

    x1 = dx1*dx1 + dy1*dy1 + dz1*dz1;
    x2 = dx2*dx2 + dy2*dy2 + dz2*dz2;

    if (root == NO) {
      if (x1 >= x2)
	*(f++) = x1;
      else
	*(f++) = x2;
    }
    else {
      if (x1 >= x2)
	*(f++) = sqrt(x1);
      else
	*(f++) = sqrt(x2);
    }
  }

  return SUCCEEDED;
}


int mesh_compute_mesh_bounding_sphere(struct object_mesh *mesh) {

  double nx, ny, nz, px, py, pz, mx, my, mz, *t, r;
  int i;


  t = mesh->vertice_list;
  nx = *(t++);
  px = nx;
  ny = *(t++);
  py = ny;
  nz = *(t++);
  pz = nz;

  /* first compute the bounding box */
  for (i = 0; i < mesh->vertices-1; i++) {
    if (*(t+0) < nx)
      nx = *(t+0);
    else if (*(t+0) > px)
      px = *(t+0);
    if (*(t+1) < ny)
      ny = *(t+1);
    else if (*(t+1) > py)
      py = *(t+1);
    if (*(t+2) < nz)
      nz = *(t+2);
    else if (*(t+2) > pz)
      pz = *(t+2);
    t += 3;
  }

  /* then create the bouding sphere surrounding the bounding box */
  mx = nx + ((px-nx)/2);
  my = ny + ((py-ny)/2);
  mz = nz + ((pz-nz)/2);
  r  = sqrt((nx-mx)*(nx-mx) + (ny-my)*(ny-my) + (nz-mz)*(nz-mz));

  /* write the sphere radius and centre */
  mesh->bounding_sphere_r = r;
  mesh->bounding_sphere[0] = mx;
  mesh->bounding_sphere[1] = my;
  mesh->bounding_sphere[2] = mz;

  /*
  r1 = 0;
  t = mesh->vertice_list;
  for (i = 0; i < mesh->vertices; i++) {
    r2 = *t* *t + (*(t+1) * *(t+1)) + (*(t+2) * *(t+2));
    if (r2 > r1)
      r1 = r2;
    t += 3;
  }

  mesh->bounding_sphere_r = sqrt(r1);
  mesh->bounding_sphere[0] = 0;
  mesh->bounding_sphere[1] = 0;
  mesh->bounding_sphere[2] = 0;
  */

  return SUCCEEDED;
}


int mesh_compute_meshes_bounding_sphere(struct object_mesh *mesh, double *sphere, double *sphere_r) {

  double nx, ny, nz, px, py, pz, mx, my, mz, *t, r;
  int i;


  t = mesh->vertice_list;
  nx = *(t++);
  px = nx;
  ny = *(t++);
  py = ny;
  nz = *(t++);
  pz = nz;

  /* first compute the bounding box */
  while (mesh != NULL) {
    t = mesh->vertice_list;
    for (i = 0; i < mesh->vertices-1; i++) {
      if (*(t+0) < nx)
	nx = *(t+0);
      else if (*(t+0) > px)
	px = *(t+0);
      if (*(t+1) < ny)
	ny = *(t+1);
      else if (*(t+1) > py)
	py = *(t+1);
      if (*(t+2) < nz)
	nz = *(t+2);
      else if (*(t+2) > pz)
	pz = *(t+2);
      t += 3;
    }
    mesh = mesh->next;
  }

  /* then create the bouding sphere surrounding the bounding box */
  mx = nx + ((px-nx)/2);
  my = ny + ((py-ny)/2);
  mz = nz + ((pz-nz)/2);
  r  = sqrt((nx-mx)*(nx-mx) + (ny-my)*(ny-my) + (nz-mz)*(nz-mz));

  /* write the sphere radius and centre */
  *sphere_r = r;
  sphere[0] = mx;
  sphere[1] = my;
  sphere[2] = mz;

  return SUCCEEDED;
}


int mesh_compute_face_normals(struct object_mesh *mesh) {

  double *t, x1, y1, z1, x2, y2, z2, x3, y3, z3, dx1, dy1, dz1, dx2, dy2, dz2;
  float *f;
  int *l, i, n;


  if (mesh == NULL)
    return SUCCEEDED;

  if (mesh->face_normal_list == NULL) {
    mesh->face_normal_list = malloc(sizeof(float) * mesh->triangles * 3);
    if (mesh->face_normal_list == NULL) {
      fprintf(stderr, "MESH_COMPUTE_FACE_NORMALS: Out of memory error.\n");
      return FAILED;
    }
  }

  /* compute triangle face normals */
  l = mesh->triangle_list;
  f = mesh->face_normal_list;
  t = mesh->vertice_list;
  for (i = 0; i < mesh->triangles; i++) {
    n = *(l++) * 3;
    x1 = t[n + 0];
    y1 = t[n + 1];
    z1 = t[n + 2];
    n = *(l++) * 3;
    x2 = t[n + 0];
    y2 = t[n + 1];
    z2 = t[n + 2];
    n = *(l++) * 3;
    x3 = t[n + 0];
    y3 = t[n + 1];
    z3 = t[n + 2];

    dx1 = x2 - x1;
    dy1 = y2 - y1;
    dz1 = z2 - z1;
    dx2 = x3 - x1;
    dy2 = y3 - y1;
    dz2 = z3 - z1;
    x1 = dy1*dz2 - dy2*dz1;
    x2 = dx2*dz1 - dx1*dz2;
    x3 = dx1*dy2 - dx2*dy1;

    dx1 = sqrt(x1*x1 + x2*x2 + x3*x3);

    *(f++) = (float)(x1/dx1);
    *(f++) = (float)(x2/dx1);
    *(f++) = (float)(x3/dx1);
  }

  return SUCCEEDED;
}


int mesh_compute_vertex_normals(struct object_mesh *mesh) {

  float *f, *v, d;
  int *l, i, n;


  if (mesh == NULL)
    return SUCCEEDED;

  if (mesh->vertex_normal_list == NULL) {
    mesh->vertex_normal_list = malloc(sizeof(float) * mesh->vertices*3);
    if (mesh->vertex_normal_list == NULL) {
      fprintf(stderr, "MESH_COMPUTE_VERTEX_NORMALS: Out of memory error.\n");
      return FAILED;
    }
  }

  v = mesh->vertex_normal_list;

  /* set all vertex normals to zero vectors */
  for (i = 0; i < mesh->vertices*3; i++)
    v[i] = 0.0;

  /* compute vertex normals */
  l = mesh->triangle_list;
  f = mesh->face_normal_list;
  for (i = 0; i < mesh->triangles; i++) {
    n = *(l++) * 3;
    v[n + 0] += f[0];
    v[n + 1] += f[1];
    v[n + 2] += f[2];
    n = *(l++) * 3;
    v[n + 0] += f[0];
    v[n + 1] += f[1];
    v[n + 2] += f[2];
    n = *(l++) * 3;
    v[n + 0] += f[0];
    v[n + 1] += f[1];
    v[n + 2] += f[2];
    f += 3;
  }

  /* normalize the vertex normals */
  for (i = 0; i < mesh->vertices; i++) {
    d = sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
    v[0] /= d;
    v[1] /= d;
    v[2] /= d;
    v += 3;
  }

  return SUCCEEDED;
}
